Imports Microsoft.VisualBasic
Imports System.Runtime.InteropServices

Public Class CToolTip
  Inherits NativeWindow

  ' ==================================================================================================
  ' CToolTip.vb
  '
  ' TimoSoft ToolTip class
  ' written by Timo "TimoSoft" Kunze (Timo@TimoSoft-Software.com)
  '
  ' This is an implementation of the Windows ToolTip control.
  '
  ' release history:
  ' ----------------
  '  1.1 (08/04/2004) Create() now takes a Control instead of an IntPtr as Parent
  '                   getting the font from the parent now works
  '  1.0 (07/30/2004) initial release
  '
  ' known issues:
  ' -------------
  '  - events are not implemented
  '
  ' --------------------------------------------------------------------------------------------------
  ' visit http://www.TimoSoft-Software.de
  ' ==================================================================================================


#Region "Definitions"
	Delegate Function WinMainDelegate(ByVal hWnd As IntPtr, ByVal Msg As Integer, ByVal wParam As Integer, ByVal lParam As Integer) As Integer

	Private Enum CommonControlsConsts
		ICC_BAR_CLASSES = &H4
	End Enum

	Private Enum ExWindowStyleConsts
		WS_EX_ACCEPTFILES = &H10
		WS_EX_APPWINDOW = &H40000
		WS_EX_CLIENTEDGE = &H200
		WS_EX_COMPOSITED = &H2000000
		WS_EX_CONTEXTHELP = &H400
		WS_EX_CONTROLPARENT = &H10000
		WS_EX_DLGMODALFRAME = &H1
		WS_EX_LAYERED = &H80000
		WS_EX_LAYOUTRTL = &H400000
		WS_EX_LEFT = &H0
		WS_EX_LEFTSCROLLBAR = &H4000
		WS_EX_LTRREADING = &H0
		WS_EX_MDICHILD = &H40
		WS_EX_NOACTIVATE = &H8000000
		WS_EX_NOINHERITLAYOUT = &H100000
		WS_EX_NOPARENTNOTIFY = &H4
		WS_EX_RIGHT = &H1000
		WS_EX_RIGHTSCROLLBAR = &H0
		WS_EX_RTLREADING = &H2000
		WS_EX_STATICEDGE = &H20000
		WS_EX_TOOLWINDOW = &H80
		WS_EX_TOPMOST = &H8
		WS_EX_TRANSPARENT = &H20
		WS_EX_WINDOWEDGE = &H100
		WS_EX_OVERLAPPEDWINDOW = (WS_EX_WINDOWEDGE Or WS_EX_CLIENTEDGE)
		WS_EX_PALETTEWINDOW = (WS_EX_WINDOWEDGE Or WS_EX_TOOLWINDOW Or WS_EX_TOPMOST)
	End Enum

	Private Enum ToolTipStyleConsts
		TTS_ALWAYSTIP = &H1
		TTS_BALLOON = &H40
		TTS_CLOSE = &H80
		TTS_NOANIMATE = &H10
		TTS_NOFADE = &H20
		TTS_NOPREFIX = &H2
	End Enum

	Private Enum WindowStyleConsts
		WS_BORDER = &H800000
		WS_CAPTION = &HC00000
		WS_CHILD = &H40000000
		WS_CLIPCHILDREN = &H2000000
		WS_CLIPSIBLINGS = &H4000000
		WS_DISABLED = &H8000000
		WS_DLGFRAME = &H400000
		WS_GROUP = &H20000
		WS_HSCROLL = &H100000
		WS_MAXIMIZE = &H1000000
		WS_MAXIMIZEBOX = &H10000
		WS_MINIMIZE = &H20000000
		WS_MINIMIZEBOX = &H20000
		WS_OVERLAPPED = &H0
		WS_POPUP = &H80000000
		WS_SYSMENU = &H80000
		WS_TABSTOP = &H10000
		WS_THICKFRAME = &H40000
		WS_VISIBLE = &H10000000
		WS_VSCROLL = &H200000
		WS_CHILDWINDOW = WS_CHILD
		WS_ICONIC = WS_MINIMIZE
		WS_OVERLAPPEDWINDOW = (WS_OVERLAPPED Or WS_CAPTION Or WS_SYSMENU Or WS_THICKFRAME Or WS_MINIMIZEBOX Or WS_MAXIMIZEBOX)
		WS_POPUPWINDOW = (WS_POPUP Or WS_BORDER Or WS_SYSMENU)
		WS_SIZEBOX = WS_THICKFRAME
		WS_TILED = WS_OVERLAPPED
		WS_TILEDWINDOW = WS_OVERLAPPEDWINDOW
	End Enum


	Public Enum BitmapPositionConstants
		bpLeft
		bpTop
		bpRight
		bpBottom
	End Enum

	Public Enum ToolTipTitleIconConstants
		tttiNone = 0
		tttiInfo = 1
		tttiWarning = 2
		tttiError = 3
	End Enum


	Private Const CLR_INVALID As Integer = -1
	Private Const LPSTR_TEXTCALLBACK As Integer = -1
	Private Const NM_FIRST As Integer = 0
	Private Const TTN_FIRST As Integer = -520
	Private Const TTN_LAST As Integer = -549
	Private Const WM_USER As Integer = &H400
	' classname for buttons
	Private Const WC_TOOLTIP As String = "tooltips_class32"

	' common controls messages
	Private Const CCM_FIRST As Integer = &H2000
	Private Const CCM_DPISCALE As Integer = (CCM_FIRST + &HC)
	Private Const CCM_GETCOLORSCHEME As Integer = (CCM_FIRST + 3)
	Private Const CCM_GETDROPTARGET As Integer = (CCM_FIRST + 4)
	Private Const CCM_GETUNICODEFORMAT As Integer = (CCM_FIRST + 6)
	Private Const CCM_GETVERSION As Integer = (CCM_FIRST + &H8)
	Private Const CCM_SETBKCOLOR As Integer = (CCM_FIRST + 1)
	Private Const CCM_SETCOLORSCHEME As Integer = (CCM_FIRST + 2)
	Private Const CCM_SETNOTIFYWINDOW As Integer = (CCM_FIRST + &H9)
	Private Const CCM_SETUNICODEFORMAT As Integer = (CCM_FIRST + 5)
	Private Const CCM_SETVERSION As Integer = (CCM_FIRST + &H7)
	Private Const CCM_SETWINDOWTHEME As Integer = (CCM_FIRST + &HB)

	' constants used with CustomDraw
	Private Const CDDS_MAPPART As Integer = &H5
	Private Const CDDS_POSTERASE As Integer = &H4
	Private Const CDDS_PREERASE As Integer = &H3
	Private Const CDDS_ITEM As Integer = &H10000
	Private Const CDDS_PREPAINT As Integer = &H1
	Private Const CDDS_POSTPAINT As Integer = &H2
	Private Const CDDS_ITEMPOSTERASE As Integer = (CDDS_ITEM Or CDDS_POSTERASE)
	Private Const CDDS_ITEMPOSTPAINT As Integer = (CDDS_ITEM Or CDDS_POSTPAINT)
	Private Const CDDS_ITEMPREERASE As Integer = (CDDS_ITEM Or CDDS_PREERASE)
	Private Const CDDS_ITEMPREPAINT As Integer = (CDDS_ITEM Or CDDS_PREPAINT)
	Private Const CDDS_SUBITEM As Integer = &H20000
	Private Const CDIS_CHECKED As Integer = &H8
	Private Const CDIS_DEFAULT As Integer = &H20
	Private Const CDIS_DISABLED As Integer = &H4
	Private Const CDIS_FOCUS As Integer = &H10
	Private Const CDIS_GRAYED As Integer = &H2
	Private Const CDIS_HOT As Integer = &H40
	Private Const CDIS_INDETERMINATE As Integer = &H100
	Private Const CDIS_MARKED As Integer = &H80
	Private Const CDIS_SELECTED As Integer = &H1
	Private Const CDIS_SHOWKEYBOARDCUES As Integer = &H200
	Private Const CDRF_DODEFAULT As Integer = &H0
	Private Const CDRF_NEWFONT As Integer = &H2
	Private Const CDRF_NOTIFYITEMDRAW As Integer = &H20
	Private Const CDRF_NOTIFYPOSTERASE As Integer = &H40
	Private Const CDRF_NOTIFYPOSTPAINT As Integer = &H10
	Private Const CDRF_NOTIFYSUBITEMDRAW As Integer = &H20
	Private Const CDRF_SKIPDEFAULT As Integer = &H4

	' tooltip messages
	Private Const TTM_ACTIVATE As Integer = (WM_USER + 1)
	Private Const TTM_ADDTOOL As Integer = (WM_USER + 50)
	Private Const TTM_DELTOOL As Integer = (WM_USER + 51)
	Private Const TTM_GETDELAYTIME As Integer = (WM_USER + 21)
	Private Const TTM_GETMARGIN As Integer = (WM_USER + 27)
	Private Const TTM_GETMAXTIPWIDTH As Integer = (WM_USER + 25)
	Private Const TTM_GETTEXT As Integer = (WM_USER + 56)
	Private Const TTM_GETTIPBKCOLOR As Integer = (WM_USER + 22)
	Private Const TTM_GETTIPTEXTCOLOR As Integer = (WM_USER + 23)
	Private Const TTM_GETTITLE As Integer = (WM_USER + 35)
	Private Const TTM_POP As Integer = (WM_USER + 28)
	Private Const TTM_SETDELAYTIME As Integer = (WM_USER + 3)
	Private Const TTM_SETMARGIN As Integer = (WM_USER + 26)
	Private Const TTM_SETMAXTIPWIDTH As Integer = (WM_USER + 24)
	Private Const TTM_SETTIPBKCOLOR As Integer = (WM_USER + 19)
	Private Const TTM_SETTIPTEXTCOLOR As Integer = (WM_USER + 20)
	Private Const TTM_SETTITLE As Integer = (WM_USER + 33)
	Private Const TTM_TRACKACTIVATE As Integer = (WM_USER + 17)
	Private Const TTM_TRACKPOSITION As Integer = (WM_USER + 18)
	Private Const TTM_UPDATETIPTEXT As Integer = (WM_USER + 57)

	' common controls notifications
	Private Const NM_CUSTOMDRAW As Integer = (NM_FIRST - 12)

	' tooltip notifications
	Private Const TTN_GETDISPINFO As Integer = (TTN_FIRST - 10)
	Private Const TTN_LINKCLICK As Integer = (TTN_FIRST - 3)
	Private Const TTN_NEEDTEXT As Integer = TTN_GETDISPINFO
	Private Const TTN_POP As Integer = (TTN_FIRST - 2)
	Private Const TTN_SHOW As Integer = (TTN_FIRST - 1)

	' tooltip flags
	Private Const TTF_CENTERTIP As Integer = &H2
	Private Const TTF_DI_SETITEM As Integer = &H8000
	Private Const TTF_IDISHWND As Integer = &H1
	Private Const TTF_PARSELINKS As Integer = &H1000
	Private Const TTF_SUBCLASS As Integer = &H10
	Private Const TTF_TRACK As Integer = &H20
	Private Const TTF_TRANSPARENT As Integer = &H100

	' constants for TTM_GETDELAYTIME
	Private Const TTDT_AUTOPOP As Integer = 2
	Private Const TTDT_INITIAL As Integer = 3
	Private Const TTDT_RESHOW As Integer = 1

	' constants used with GetWindowLongPtr()
	Private Const GWL_EXSTYLE As Integer = (-20)
	Private Const GWL_STYLE As Integer = (-16)
	Private Const GWLP_WNDPROC As Integer = (-4)

	' constants used with ShowWindow()
	Private Const SW_HIDE As Integer = 0

	<System.Runtime.InteropServices.StructLayout(LayoutKind.Sequential)> _
	Private Structure BITMAPData
		Public bmType As Integer
		Public bmWidth As Integer
		Public bmHeight As Integer
		Public bmWidthBytes As Integer
		Public bmPlanes As Short
		Public bmBitsPixel As Short
		Public bmBits As IntPtr
	End Structure

	<StructLayout(LayoutKind.Sequential)> _
	Private Structure RECT
		Public left As Integer
		Public top As Integer
		Public right As Integer
		Public bottom As Integer
	End Structure

	<System.Runtime.InteropServices.StructLayout(LayoutKind.Sequential)> _
	Private Structure NMHDR
		Public hWndFrom As IntPtr
		Public idfrom As Integer
		Public code As Integer
	End Structure

	<System.Runtime.InteropServices.StructLayout(LayoutKind.Sequential)> _
	Private Structure NMCUSTOMDRAW
		Public hdr As NMHDR
		Public dwDrawStage As Integer
		Public hDC As IntPtr
		Public rc As RECT
		Public dwItemSpec As Integer
		Public uItemState As Integer
		Public lItemlParam As IntPtr
	End Structure

	<System.Runtime.InteropServices.StructLayout(LayoutKind.Sequential)> _
	Private Structure NMTTCUSTOMDRAW
		Public nmcd As NMCUSTOMDRAW
		Public uDrawFlags As Integer
	End Structure

	<System.Runtime.InteropServices.StructLayout(LayoutKind.Sequential, Pack:=4)> _
	Private Structure PAINTSTRUCT
		Public hDC As IntPtr
		Public fErase As Boolean
		Public rcPaint As RECT
		Public fRestore As Boolean
		Public fIncUpdate As Boolean
		<MarshalAs(UnmanagedType.ByValArray, SizeConst:=32)> _
		Public rgbReserved() As Byte
	End Structure

	<System.Runtime.InteropServices.StructLayout(LayoutKind.Sequential)> _
	Private Structure TOOLINFO
		Public cbSize As Integer
		Public uFlags As Integer
		Public hWnd As IntPtr
		Public uId As IntPtr
		Public rc As RECT
		Public hinst As IntPtr
		<MarshalAs(UnmanagedType.LPTStr)> _
		Public lpszText As String
		Public lParam As Integer
		'Public lpReserved As Integer     ' needs Windows XP or higher
	End Structure

	<System.Runtime.InteropServices.StructLayout(LayoutKind.Sequential)> _
	Private Structure INITCOMMONCONTROLSEXData
		Public dwSize As Integer
		Public dwICC As Integer
	End Structure

	<System.Runtime.InteropServices.StructLayout(LayoutKind.Sequential)> _
	Private Structure NMTTDISPINFO
		Public hdr As NMHDR
		<MarshalAs(UnmanagedType.LPTStr)> _
		Public lpszText As String
		<MarshalAs(UnmanagedType.LPArray, SizeConst:=80)> _
		Public szText() As Char
		Public hinst As IntPtr
		Public uFlags As Integer
		Public lParam As Integer
	End Structure

	<System.Runtime.InteropServices.StructLayout(LayoutKind.Sequential)> _
	Private Structure TTGETTITLE
		Public dwSize As Integer
		Public uTitleBitmap As Integer
		Public cch As Integer
		<MarshalAs(UnmanagedType.LPWStr)> _
		Public pszTitle As String
	End Structure


	Private bAttached As Boolean
	Private BMPSize As Size
	Private oldWinMainParent As IntPtr
	Private rcContent As RECT
	Private TextSize As Size

	Private propAlwaysTip As Boolean
	Private propAutoPopDelayTime As Short
	Private propBackColor As Color
	Private propBalloonStyle As Boolean
	Private propBitmapMarginX As Integer
	Private propBitmapMarginY As Integer
	Private propBitmapPosition As BitmapPositionConstants
	Private prophBitmap As IntPtr
	Private propFontToUse As Font
	Private prophWndParent As IntPtr
	Private propInitialDelayTime As Short
	Private propMargin As Rectangle
	Private propMaskColor As Color
	Private propMaxTipWidth As Integer
	Private propNoAnimation As Boolean
	Private propNoFade As Boolean
	Private propReshowDelayTime As Short
	Private propShowCloseButton As Boolean
	Private propText As String
	Private propTextColor As Color
	Private propTitleIcon As ToolTipTitleIconConstants
	Private propTitleText As String
	Private propUseMaskColor As Boolean


	Private Declare Unicode Function BeginPaint Lib "user32.dll" (ByVal hWnd As IntPtr, ByRef lpPaint As PAINTSTRUCT) As IntPtr
	Private Declare Unicode Function CallWindowProc Lib "user32.dll" Alias "CallWindowProcW" (ByVal adr As IntPtr, ByVal hWnd As IntPtr, ByVal Msg As Integer, ByVal wParam As Integer, ByVal lParam As Integer) As Integer
	Private Declare Unicode Function CreateBitmap Lib "gdi32.dll" (ByVal cx As Integer, ByVal cy As Integer, ByVal ClrPlanes As Integer, ByVal BPP As Integer, ByVal Bits As IntPtr) As IntPtr
	Private Declare Unicode Function CreateCompatibleDC Lib "gdi32.dll" (ByVal hDC As IntPtr) As IntPtr
	Private Declare Unicode Function CreateCompatibleBitmap Lib "gdi32.dll" (ByVal hDC As IntPtr, ByVal Width As Integer, ByVal Height As Integer) As IntPtr
	Private Declare Unicode Function CreateSolidBrush Lib "gdi32.dll" (ByVal Clr As Integer) As IntPtr
	Private Declare Unicode Function CreateWindowEx Lib "user32.dll" Alias "CreateWindowExW" (ByVal StyleEx As Integer, ByVal Klasse As String, ByVal Caption As String, ByVal Style As Integer, ByVal x As Integer, ByVal y As Integer, ByVal Width As Integer, ByVal Height As Integer, ByVal hWnd_Parent As IntPtr, ByVal hMenu As IntPtr, ByVal hInstanz As IntPtr, ByVal lParam As IntPtr) As IntPtr
	Private Declare Unicode Function DeleteDC Lib "gdi32.dll" (ByVal hDC As IntPtr) As Integer
	Private Declare Unicode Function DeleteObject Lib "gdi32.dll" (ByVal hObject As IntPtr) As Integer
	Private Declare Unicode Function DestroyWindow Lib "user32.dll" Alias "DestroyWindowW" (ByVal hWnd As IntPtr) As Integer
	Private Declare Unicode Function EndPaint Lib "user32.dll" (ByVal hWnd As IntPtr, ByRef lpPaint As PAINTSTRUCT) As IntPtr
	Private Declare Unicode Function FillRect Lib "user32.dll" (ByVal hDC As IntPtr, ByRef rc As RECT, ByVal hBrush As IntPtr) As Integer
	Private Declare Unicode Function GetClientRect Lib "user32.dll" (ByVal hWnd As IntPtr, ByRef rc As RECT) As Integer
	Private Declare Unicode Function GetObject Lib "gdi32.dll" Alias "GetObjectW" (ByVal hObject As IntPtr, ByVal Buffersize As Integer, ByRef Data As BITMAPData) As IntPtr
	Private Declare Unicode Function GetParent Lib "user32.dll" (ByVal hWnd As IntPtr) As IntPtr
	Private Declare Unicode Function GetWindowLongPtr Lib "user32.dll" Alias "GetWindowLongW" (ByVal hWnd As IntPtr, ByVal Prop As Integer) As Integer
	Private Declare Unicode Function InitCommonControlsEx Lib "comctl32.dll" (ByRef Data As INITCOMMONCONTROLSEXData) As Integer
	Private Declare Unicode Function lstrcpyn Lib "kernel32.dll" Alias "lstrcpynW" (ByVal pTarget As IntPtr, ByVal Src As String, ByVal MaxLength As Integer) As Integer
	Private Declare Unicode Function PostMessage Lib "user32.dll" Alias "PostMessageW" (ByVal hWnd As IntPtr, ByVal Msg As Integer, ByVal wParam As IntPtr, ByVal lParam As IntPtr) As Integer
	Private Declare Unicode Function PostMessage Lib "user32.dll" Alias "PostMessageW" (ByVal hWnd As IntPtr, ByVal Msg As Integer, ByVal wParam As Integer, ByVal lParam As IntPtr) As Integer
	Private Declare Unicode Function PostMessage Lib "user32.dll" Alias "PostMessageW" (ByVal hWnd As IntPtr, ByVal Msg As Integer, ByVal wParam As IntPtr, ByVal lParam As Integer) As Integer
	Private Declare Unicode Function PostMessage Lib "user32.dll" Alias "PostMessageW" (ByVal hWnd As IntPtr, ByVal Msg As Integer, ByVal wParam As Integer, ByVal lParam As Integer) As Integer
	Private Declare Unicode Function SelectObject Lib "gdi32.dll" (ByVal hDC As IntPtr, ByVal hObject As IntPtr) As IntPtr
	Private Declare Unicode Function SendMessage Lib "user32.dll" Alias "SendMessageW" (ByVal hWnd As IntPtr, ByVal Msg As Integer, ByVal wParam As IntPtr, ByVal lParam As IntPtr) As IntPtr
	Private Declare Unicode Function SendMessage Lib "user32.dll" Alias "SendMessageW" (ByVal hWnd As IntPtr, ByVal Msg As Integer, ByVal wParam As Integer, ByVal lParam As IntPtr) As IntPtr
	Private Declare Unicode Function SendMessage Lib "user32.dll" Alias "SendMessageW" (ByVal hWnd As IntPtr, ByVal Msg As Integer, ByVal wParam As IntPtr, ByVal lParam As Integer) As IntPtr
	Private Declare Unicode Function SendMessage Lib "user32.dll" Alias "SendMessageW" (ByVal hWnd As IntPtr, ByVal Msg As Integer, ByVal wParam As Integer, ByVal lParam As Integer) As IntPtr
	Private Declare Unicode Function SendMessage Lib "user32.dll" Alias "SendMessageW" (ByVal hWnd As IntPtr, ByVal Msg As Integer, ByVal wParam As Integer, ByRef lParam As TOOLINFO) As IntPtr
	Private Declare Unicode Function SendMessage Lib "user32.dll" Alias "SendMessageW" (ByVal hWnd As IntPtr, ByVal Msg As Integer, ByVal wParam As Integer, ByRef lParam As TTGETTITLE) As IntPtr
	Private Declare Unicode Function SendMessage Lib "user32.dll" Alias "SendMessageW" (ByVal hWnd As IntPtr, ByVal Msg As Integer, ByVal wParam As Integer, ByRef lParam As Rectangle) As IntPtr
	Private Declare Unicode Function SetBkColor Lib "gdi32.dll" (ByVal hDC As IntPtr, ByVal Clr As Color) As Color
	Private Declare Unicode Function SetParent Lib "user32.dll" Alias "SetParentW" (ByVal hWndChild As IntPtr, ByVal hWndNewParent As IntPtr) As IntPtr
	Private Declare Unicode Function SetWindowLongPtr Lib "user32.dll" Alias "SetWindowLongW" (ByVal hWnd As IntPtr, ByVal Prop As Integer, ByVal NewProp As Integer) As Integer
	Private Declare Unicode Function SetWindowLongPtr Lib "user32.dll" Alias "SetWindowLongW" (ByVal hWnd As IntPtr, ByVal Prop As Integer, ByVal NewProp As WinMainDelegate) As IntPtr
	Private Declare Unicode Function SetWindowLongPtr Lib "user32.dll" Alias "SetWindowLongW" (ByVal hWnd As IntPtr, ByVal Prop As Integer, ByVal NewProp As IntPtr) As IntPtr
	Private Declare Unicode Function ShowWindow Lib "user32.dll" Alias "ShowWindowW" (ByVal hWnd As IntPtr, ByVal Flags As Integer) As Integer


	Public Event BeforeHide()
	Public Event BeforeShow()
	Public Event ClickedLink()
	Public Event NeedText(ByRef TextToDisplay As String, ByRef StoreData As Boolean)
#End Region


  Public Sub New()
    MyBase.New()

    Dim Data As INITCOMMONCONTROLSEXData
    Dim ptr As Integer

    Data.dwICC = CommonControlsConsts.ICC_BAR_CLASSES
    Data.dwSize = Marshal.SizeOf(Data)
    InitCommonControlsEx(Data)

    propAlwaysTip = True
    propAutoPopDelayTime = 5000
    propBackColor = SystemColors.Info
    propBalloonStyle = False
		propBitmapMarginX = 7
		propBitmapMarginY = 7
		propBitmapPosition = BitmapPositionConstants.bpLeft
		propInitialDelayTime = 500
		propMaskColor = Color.Magenta
		propMaxTipWidth = 300
    propNoAnimation = False
    propNoFade = False
    propReshowDelayTime = 100
    propShowCloseButton = False
    propTextColor = SystemColors.InfoText
    propTitleIcon = ToolTipTitleIconConstants.tttiNone
    propTitleText = ""
		propUseMaskColor = True
	End Sub


	Public Property AlwaysTip() As Boolean
		Get
			Dim Style As Integer

			If Not Me.Handle.Equals(IntPtr.Zero) Then
				Style = GetWindowLongPtr(Me.Handle, GWL_STYLE)
				propAlwaysTip = ((Style And ToolTipStyleConsts.TTS_ALWAYSTIP) = ToolTipStyleConsts.TTS_ALWAYSTIP)
			End If
			Return propAlwaysTip
		End Get
		Set(ByVal Value As Boolean)
			Dim Style As Integer

			propAlwaysTip = Value
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				Style = GetWindowLongPtr(Me.Handle, GWL_STYLE)
				If propAlwaysTip Then
					Style = Style Or ToolTipStyleConsts.TTS_ALWAYSTIP
				Else
					Style = Style And Not ToolTipStyleConsts.TTS_ALWAYSTIP
				End If
				SetWindowLongPtr(Me.Handle, GWL_STYLE, Style)
			End If
		End Set
	End Property

	Public Property AutoPopDelayTime() As Short
		Get
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				propAutoPopDelayTime = CShort(SendMessage(Me.Handle, TTM_GETDELAYTIME, TTDT_AUTOPOP, 0).ToInt32)
			End If
			Return propAutoPopDelayTime
		End Get
		Set(ByVal Value As Short)
			propAutoPopDelayTime = Value
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				SendMessage(Me.Handle, TTM_SETDELAYTIME, TTDT_AUTOPOP, MAKELONG(propAutoPopDelayTime, 0))
			End If
		End Set
	End Property

	Public Property BackColor() As Color
		Get
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				propBackColor = RGBColorToColor(SendMessage(Me.Handle, TTM_GETTIPBKCOLOR, 0, 0).ToInt32)
			End If
			Return propBackColor
		End Get
		Set(ByVal Value As Color)
			propBackColor = Value
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				PostMessage(Me.Handle, TTM_SETTIPBKCOLOR, ColorToRGBColor(propBackColor), 0)
			End If
		End Set
	End Property

	Public Property BalloonStyle() As Boolean
		Get
			Dim Style As Integer

			If Not Me.Handle.Equals(IntPtr.Zero) Then
				Style = GetWindowLongPtr(Me.Handle, GWL_STYLE)
				propBalloonStyle = ((Style And ToolTipStyleConsts.TTS_BALLOON) = ToolTipStyleConsts.TTS_BALLOON)
			End If
			Return propBalloonStyle
		End Get
		Set(ByVal Value As Boolean)
			Dim Style As Integer

			propBalloonStyle = Value
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				Style = GetWindowLongPtr(Me.Handle, GWL_STYLE)
				If propBalloonStyle Then
					Style = Style Or ToolTipStyleConsts.TTS_BALLOON
					Style = Style And Not WindowStyleConsts.WS_BORDER
				Else
					Style = Style And Not ToolTipStyleConsts.TTS_BALLOON
					Style = Style Or WindowStyleConsts.WS_BORDER
				End If
				SetWindowLongPtr(Me.Handle, GWL_STYLE, Style)
			End If
		End Set
	End Property

	Public Property BitmapMarginX() As Integer
		Get
			Return propBitmapMarginX
		End Get
		Set(ByVal Value As Integer)
			propBitmapMarginX = Value
		End Set
	End Property

	Public Property BitmapMarginY() As Integer
		Get
			Return propBitmapMarginY
		End Get
		Set(ByVal Value As Integer)
			propBitmapMarginY = Value
		End Set
	End Property

	Public Property BitmapPosition() As BitmapPositionConstants
		Get
			Return propBitmapPosition
		End Get
		Set(ByVal Value As BitmapPositionConstants)
			propBitmapPosition = Value
		End Set
	End Property

	Public Property FontToUse() As Font
		Get
			Const WM_GETFONT As Integer = &H31

			If Not Me.Handle.Equals(IntPtr.Zero) Then
				propFontToUse = Font.FromHfont(SendMessage(Me.Handle, WM_GETFONT, 0, 0))
			End If
			Return propFontToUse
		End Get
		Set(ByVal Value As Font)
			Const WM_SETFONT As Integer = &H30

			propFontToUse = Value
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				PostMessage(Me.Handle, WM_SETFONT, propFontToUse.ToHfont(), 1)
			End If
		End Set
	End Property

	Public Property hBitmap() As IntPtr
		Get
			Return prophBitmap
		End Get
		Set(ByVal Value As IntPtr)
			Dim BMPData As BITMAPData

			hBitmap = Value
			GetObject(hBitmap, Marshal.SizeOf(BMPData), BMPData)
			BMPSize.Height = BMPData.bmHeight
			BMPSize.Width = BMPData.bmWidth
		End Set
	End Property

	Public ReadOnly Property hWnd() As IntPtr
		Get
			Return Me.Handle
		End Get
	End Property

	Public Property hWndParent() As IntPtr
		Get
			If Not Me.Handle.Equals(IntPtr.Zero) Then prophWndParent = GetParent(Me.Handle)
			Return prophWndParent
		End Get
		Set(ByVal Value As IntPtr)
			prophWndParent = Value
			If Not Me.Handle.Equals(IntPtr.Zero) Then SetParent(Me.Handle, prophWndParent)
			SubclassParent()
		End Set
	End Property

	Public Property InitialDelayTime() As Short
		Get
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				propInitialDelayTime = CShort(SendMessage(Me.Handle, TTM_GETDELAYTIME, TTDT_INITIAL, 0).ToInt32)
			End If
			Return propInitialDelayTime
		End Get
		Set(ByVal Value As Short)
			propInitialDelayTime = Value
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				SendMessage(Me.Handle, TTM_SETDELAYTIME, TTDT_INITIAL, MAKELONG(propInitialDelayTime, 0))
			End If
		End Set
	End Property

	Public Property Margin() As Rectangle
		Get
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				SendMessage(Me.Handle, TTM_GETMARGIN, 0, propMargin)
			End If
			Return New Rectangle(propMargin.Left, propMargin.Top, propMargin.Left + propMargin.Right, propMargin.Top + propMargin.Bottom)
		End Get
		Set(ByVal Value As Rectangle)
			propMargin = Value
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				SendMessage(Me.Handle, TTM_SETMARGIN, 0, propMargin)
			End If
		End Set
	End Property

	Public Property MaskColor() As Color
		Get
			Return propMaskColor
		End Get
		Set(ByVal Value As Color)
			propMaskColor = Value
		End Set
	End Property

	Public Property MaxTipWidth() As Integer
		Get
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				propMaxTipWidth = SendMessage(Me.Handle, TTM_GETMAXTIPWIDTH, 0, 0).ToInt32
			End If
			Return propMaxTipWidth
		End Get
		Set(ByVal Value As Integer)
			propMaxTipWidth = Value
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				SendMessage(Me.Handle, TTM_SETMAXTIPWIDTH, 0, propMaxTipWidth)
			End If
		End Set
	End Property

	Public Property NoAnimation() As Boolean
		Get
			Dim Style As Integer

			If Not Me.Handle.Equals(IntPtr.Zero) Then
				Style = GetWindowLongPtr(Me.Handle, GWL_STYLE)
				propNoAnimation = ((Style And ToolTipStyleConsts.TTS_NOANIMATE) = ToolTipStyleConsts.TTS_NOANIMATE)
			End If
			Return propNoAnimation
		End Get
		Set(ByVal Value As Boolean)
			Dim Style As Integer

			propNoAnimation = Value
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				Style = GetWindowLongPtr(Me.Handle, GWL_STYLE)
				If propNoAnimation Then
					Style = Style Or ToolTipStyleConsts.TTS_NOANIMATE
				Else
					Style = Style And Not ToolTipStyleConsts.TTS_NOANIMATE
				End If
				SetWindowLongPtr(Me.Handle, GWL_STYLE, Style)
			End If
		End Set
	End Property

	Public Property NoFade() As Boolean
		Get
			Dim Style As Integer

			If Not Me.Handle.Equals(IntPtr.Zero) Then
				Style = GetWindowLongPtr(Me.Handle, GWL_STYLE)
				propNoFade = ((Style And ToolTipStyleConsts.TTS_NOFADE) = ToolTipStyleConsts.TTS_NOFADE)
			End If
			Return propNoFade
		End Get
		Set(ByVal Value As Boolean)
			Dim Style As Integer

			propNoFade = Value
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				Style = GetWindowLongPtr(Me.Handle, GWL_STYLE)
				If propNoFade Then
					Style = Style Or ToolTipStyleConsts.TTS_NOFADE
				Else
					Style = Style And Not ToolTipStyleConsts.TTS_NOFADE
				End If
				SetWindowLongPtr(Me.Handle, GWL_STYLE, Style)
			End If
		End Set
	End Property

	Public Property ReshowDelayTime() As Short
		Get
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				propReshowDelayTime = CShort(SendMessage(Me.Handle, TTM_GETDELAYTIME, TTDT_RESHOW, 0).ToInt32)
			End If
			Return propReshowDelayTime
		End Get
		Set(ByVal Value As Short)
			propReshowDelayTime = Value
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				SendMessage(Me.Handle, TTM_SETDELAYTIME, TTDT_RESHOW, MAKELONG(propReshowDelayTime, 0))
			End If
		End Set
	End Property

	Public Property ShowCloseButton() As Boolean
		Get
			Dim Style As Integer

			If Not Me.Handle.Equals(IntPtr.Zero) Then
				Style = GetWindowLongPtr(Me.Handle, GWL_STYLE)
				propShowCloseButton = ((Style And ToolTipStyleConsts.TTS_CLOSE) = ToolTipStyleConsts.TTS_CLOSE)
			End If
			Return propShowCloseButton
		End Get
		Set(ByVal Value As Boolean)
			Dim Style As Integer

			propShowCloseButton = Value
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				Style = GetWindowLongPtr(Me.Handle, GWL_STYLE)
				If propShowCloseButton Then
					Style = Style Or ToolTipStyleConsts.TTS_CLOSE
				Else
					Style = Style And Not ToolTipStyleConsts.TTS_CLOSE
				End If
				SetWindowLongPtr(Me.Handle, GWL_STYLE, Style)
			End If
		End Set
	End Property

	Public Property Text(ByVal hWndTool As IntPtr) As String
		Get
			Dim Data As TOOLINFO

			If Not Me.Handle.Equals(IntPtr.Zero) Then
				With Data
					.cbSize = Marshal.SizeOf(Data)
					.hinst = Process.GetCurrentProcess().Handle
					.hWnd = Me.hWndParent
					propText = Space(2048)
					.lpszText = propText
					.uId = hWndTool
					.uFlags = TTF_IDISHWND
				End With
				SendMessage(Me.Handle, TTM_GETTEXT, 0, Data)
				propText = removeNull(propText)
			End If
			Return propText
		End Get
		Set(ByVal Value As String)
			Dim Data As TOOLINFO

			propText = Value
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				With Data
					.cbSize = Marshal.SizeOf(Data)
					.hinst = Process.GetCurrentProcess().Handle
					.hWnd = Me.hWndParent
					.lpszText = propText
					.uId = hWndTool
					.uFlags = TTF_IDISHWND
				End With
				SendMessage(Me.Handle, TTM_UPDATETIPTEXT, 0, Data)
			End If
		End Set
	End Property

	Public Property TextColor() As Color
		Get
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				propTextColor = RGBColorToColor(SendMessage(Me.Handle, TTM_GETTIPTEXTCOLOR, 0, 0).ToInt32)
			End If
			Return propTextColor
		End Get
		Set(ByVal Value As Color)
			propTextColor = Value
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				PostMessage(Me.Handle, TTM_SETTIPTEXTCOLOR, ColorToRGBColor(propTextColor), 0)
			End If
		End Set
	End Property

	Public Property TitleIcon() As ToolTipTitleIconConstants
		Get
			Dim Data As TTGETTITLE
			Dim str As String

			If Not Me.Handle.Equals(IntPtr.Zero) Then
				With Data
					.dwSize = Marshal.SizeOf(Data)
					str = Space(1024)
					.pszTitle = str
					.cch = Len(str)

					SendMessage(Me.Handle, TTM_GETTITLE, 0, Data)
					propTitleIcon = CType(.uTitleBitmap, ToolTipTitleIconConstants)
				End With
			End If
			Return propTitleIcon
		End Get
		Set(ByVal Value As ToolTipTitleIconConstants)
			propTitleIcon = Value
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				SendMessage(Me.Handle, TTM_SETTITLE, propTitleIcon, Marshal.StringToHGlobalAuto(propTitleText))
			End If
		End Set
	End Property

	Public Property TitleText() As String
		Get
			Dim Data As TTGETTITLE
			Dim str As String

			If Not Me.Handle.Equals(IntPtr.Zero) Then
				With Data
					.dwSize = Marshal.SizeOf(Data)
					str = Space(1024)
					.pszTitle = str
					.cch = Len(str)

					SendMessage(Me.Handle, TTM_GETTITLE, 0, Data)
					propTitleText = removeNull(str)
				End With
			End If
			Return propTitleText
		End Get
		Set(ByVal Value As String)
			propTitleText = Value
			If Not Me.Handle.Equals(IntPtr.Zero) Then
				SendMessage(Me.Handle, TTM_SETTITLE, propTitleIcon, Marshal.StringToHGlobalAuto(propTitleText))
			End If
		End Set
	End Property

	Public Property UseMaskColor() As Boolean
		Get
			Return propUseMaskColor
		End Get
		Set(ByVal Value As Boolean)
			propUseMaskColor = Value
		End Set
	End Property


	Public Sub Activate()
		If Not Me.Handle.Equals(IntPtr.Zero) Then
			SendMessage(Me.Handle, TTM_ACTIVATE, 1, 0)
		End If
	End Sub

	Public Sub AddTool(ByVal hWndTool As IntPtr, ByVal TextToDisplay As String, Optional ByVal SubClassTool As Boolean = True, Optional ByVal Transparent As Boolean = True, Optional ByVal AdjustToTool As Boolean = True, Optional ByVal ParseLinks As Boolean = True, Optional ByVal Track As Boolean = False)
		Dim Data As TOOLINFO

		If Not Me.Handle.Equals(IntPtr.Zero) Then
			With Data
				.cbSize = Marshal.SizeOf(Data)
				.hinst = Process.GetCurrentProcess().Handle
				.hWnd = Me.hWndParent
				.lpszText = TextToDisplay
				.uId = hWndTool
				.uFlags = TTF_IDISHWND
				If AdjustToTool Then .uFlags = .uFlags Or TTF_CENTERTIP
				If ParseLinks Then .uFlags = .uFlags Or TTF_PARSELINKS
				If SubClassTool Then .uFlags = .uFlags Or TTF_SUBCLASS
				If Track Then .uFlags = .uFlags Or TTF_TRACK
				If Transparent Then .uFlags = .uFlags Or TTF_TRANSPARENT
			End With
			SendMessage(Me.Handle, TTM_ADDTOOL, 0, Data)
		End If
	End Sub

	' attaches this object to an existing tooltip control
	Public Sub Attach(ByVal hWndToolTip As IntPtr)
		Me.Destroy()
		Me.AssignHandle(hWndToolTip)
		bAttached = True
		SubclassParent()
	End Sub

	' creates a tooltip
	Public Sub Create(ByVal Parent As Control, Optional ByVal FontToUse As Font = Nothing)
		Const CW_USEDEFAULT As Integer = &H80000000
		Const WM_GETFONT As Integer = &H31
		Const WM_SETFONT As Integer = &H30
		Dim ExStyle As ExWindowStyleConsts
		Dim Style As WindowStyleConsts
		Dim TTStyle As ToolTipStyleConsts

		If Parent Is Nothing Then Exit Sub

		' you'll never know...
		Me.Destroy()
		If Me.Handle.Equals(IntPtr.Zero) Then
			prophWndParent = Parent.Handle
			If Not (FontToUse Is Nothing) Then
				Me.FontToUse = FontToUse
			ElseIf Me.FontToUse Is Nothing Then
				Me.FontToUse = Parent.Font
			End If

			ExStyle = ExWindowStyleConsts.WS_EX_LEFT Or ExWindowStyleConsts.WS_EX_LTRREADING Or ExWindowStyleConsts.WS_EX_TOOLWINDOW
			Style = WindowStyleConsts.WS_POPUP Or WindowStyleConsts.WS_VISIBLE
			TTStyle = ToolTipStyleConsts.TTS_NOPREFIX
			If propAlwaysTip Then TTStyle = TTStyle Or ToolTipStyleConsts.TTS_ALWAYSTIP
			If propBalloonStyle Then TTStyle = TTStyle Or ToolTipStyleConsts.TTS_BALLOON
			If propNoAnimation Then TTStyle = TTStyle Or ToolTipStyleConsts.TTS_NOANIMATE
			If propNoFade Then TTStyle = TTStyle Or ToolTipStyleConsts.TTS_NOFADE
			If propShowCloseButton Then TTStyle = TTStyle Or ToolTipStyleConsts.TTS_CLOSE

			' create the window
			Me.AssignHandle(CreateWindowEx(ExStyle, WC_TOOLTIP, "", Style Or TTStyle, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, Me.hWndParent, IntPtr.Zero, Process.GetCurrentProcess().Handle, IntPtr.Zero))
			If Me.Handle.Equals(IntPtr.Zero) Then Return
		End If

		' set the properties...
		PostMessage(Me.Handle, TTM_SETMAXTIPWIDTH, 0, propMaxTipWidth)
		PostMessage(Me.Handle, TTM_SETDELAYTIME, TTDT_AUTOPOP, MAKELONG(propAutoPopDelayTime, 0))
		PostMessage(Me.Handle, TTM_SETDELAYTIME, TTDT_INITIAL, MAKELONG(propInitialDelayTime, 0))
		PostMessage(Me.Handle, TTM_SETDELAYTIME, TTDT_RESHOW, MAKELONG(propReshowDelayTime, 0))
		SendMessage(Me.Handle, TTM_SETTITLE, propTitleIcon, Marshal.StringToHGlobalAuto(propTitleText))

		' ...and colors...
		If ColorToRGBColor(propBackColor) <> ColorToRGBColor(SystemColors.Info) Then
			PostMessage(Me.Handle, TTM_SETTIPBKCOLOR, ColorToRGBColor(propBackColor), 0)
		End If
		If ColorToRGBColor(propTextColor) <> ColorToRGBColor(SystemColors.InfoText) Then
			PostMessage(Me.Handle, TTM_SETTIPTEXTCOLOR, ColorToRGBColor(propTextColor), 0)
		End If

		' ...and fonts
		PostMessage(Me.Handle, WM_SETFONT, propFontToUse.ToHfont, 1)
		SubclassParent()
	End Sub

	Public Sub Deactivate()
		If Not Me.Handle.Equals(IntPtr.Zero) Then
			SendMessage(Me.Handle, TTM_ACTIVATE, 0, 0)
		End If
	End Sub

	Public Sub DelTool(ByVal hWndTool As IntPtr)
		Dim Data As TOOLINFO

		If Not Me.Handle.Equals(IntPtr.Zero) Then
			With Data
				.cbSize = Marshal.SizeOf(Data)
				.hWnd = Me.hWndParent
				.uId = hWndTool
			End With
			SendMessage(Me.Handle, TTM_DELTOOL, 0, Data)
		End If
	End Sub

	' destroys the frame
	Public Sub Destroy()
		UnSubclassParent()
		If Not Me.Handle.Equals(IntPtr.Zero) Then
			If Not bAttached Then
				ShowWindow(Me.Handle, SW_HIDE)
				SetParent(Me.Handle, IntPtr.Zero)

				Me.DestroyHandle()
			End If
		End If
	End Sub

	' detaches this object from the tooltip control (must have been previously attached)
	Public Sub Detach()
		UnSubclassParent()
		If bAttached Then
			Me.ReleaseHandle()

			bAttached = False
		End If
	End Sub

	Public Sub Hide(ByVal hWndTool As IntPtr)
		Dim Data As TOOLINFO

		If Not Me.Handle.Equals(IntPtr.Zero) Then
			If Not hWndTool.Equals(IntPtr.Zero) Then
				With Data
					.cbSize = Marshal.SizeOf(Data)
					.hWnd = Me.hWndParent
					.uId = hWndTool
					.uFlags = TTF_IDISHWND
				End With
				SendMessage(Me.Handle, TTM_TRACKACTIVATE, 0, Data)
			Else
				SendMessage(Me.Handle, TTM_POP, 0, 0)
			End If
		End If
	End Sub

	Public Sub Show(ByVal hWndTool As IntPtr)
		Dim Data As TOOLINFO

		If Not Me.Handle.Equals(IntPtr.Zero) Then
			With Data
				.cbSize = Marshal.SizeOf(Data)
				.hWnd = Me.hWndParent
				.uId = hWndTool
				.uFlags = TTF_IDISHWND
			End With
			SendMessage(Me.Handle, TTM_TRACKACTIVATE, 1, Data)
		End If
	End Sub

	Public Sub TrackTo(ByVal x As Short, ByVal y As Short)
		If Not Me.Handle.Equals(IntPtr.Zero) Then
			PostMessage(Me.Handle, TTM_TRACKPOSITION, 0, MAKELONG(x, y))
		End If
	End Sub


	Private Function ColorToRGBColor(ByVal clr As Color) As Integer
		Return (clr.R Or (CInt(clr.G) << 8) Or (CInt(clr.B) << 16))
	End Function

	Private Function RGBColorToColor(ByVal RGB As Integer) As Color
		Return Color.FromArgb(RGB Or &HFF000000)
	End Function

	' concatenates <Lo> and <Hi> to a 32-bit-number
	Private Function MAKELONG(ByVal Lo As Int16, ByVal Hi As Int16) As Integer
		Return (CInt(Lo) And &HFFFF) Or (CInt(Hi) And &HFFFF << 16)
	End Function

	' cuts <txt> at the first occurance of Chr(0)
	Private Function removeNull(ByVal txt As String) As String
		Return txt.TrimEnd(Chr(0))
	End Function

	Private Sub SubclassParent()
		'oldWinMainParent = SetWindowLongPtr(prophWndParent, GWLP_WNDPROC, AddressOf WinMain)
	End Sub

	Private Sub UnSubclassParent()
		'If Not oldWinMainParent.Equals(IntPtr.Zero) Then
		'	SetWindowLongPtr(prophWndParent, GWLP_WNDPROC, oldWinMainParent)
		'	oldWinMainParent = IntPtr.Zero
		'End If
	End Sub

	Private Function VarPtr(ByVal o As Object) As IntPtr
		Dim GC As GCHandle = GCHandle.Alloc(o, GCHandleType.Pinned)
		Dim ret As IntPtr = GC.AddrOfPinnedObject

		GC.Free()
		Return ret
	End Function

	Protected Function WinMain(ByVal hWnd As IntPtr, ByVal Msg As Integer, ByVal wParam As Integer, ByVal lParam As Integer) As Integer
		Const WM_NOTIFY As Integer = &H4E
		Dim callOldWinMain As Boolean = True
		Dim ret As Integer

		Select Case Msg
			Case WM_NOTIFY
				MsgBox("a")
		End Select

		If callOldWinMain Then
			ret = CallWindowProc(oldWinMainParent, hWnd, Msg, wParam, lParam)
		End If

		WinMain = ret
	End Function
End Class
