Imports Microsoft.VisualBasic
Imports System.Drawing
Imports System.Runtime.InteropServices

Namespace ListBoxToolTips
  Public Class ToolTipListBox
    Inherits System.Windows.Forms.NativeWindow

    Private bAdjustTools As Boolean = False
    Private bAutoGetIndex As Boolean = True
    Private bUseTimer As Boolean = False
    Private tt As New CToolTip
    Private iLastIndex As Short = -1

    Private m_ExternalTT As CToolTip
    Private m_HotItem As Integer = -1
    Private WithEvents m_Timer As New Timer

#Region "Unmanaged Code"
    <StructLayout(LayoutKind.Sequential)> _
    Public Structure POINT
      Public x As Integer
      Public y As Integer

      Public Function ToInt32() As Integer
        Return (x And &HFFFF) Or (y And &HFFFF << 16)
      End Function
    End Structure

    <StructLayout(LayoutKind.Sequential)> _
    Private Structure RECT
      Public left As Integer
      Public top As Integer
      Public right As Integer
      Public bottom As Integer
    End Structure

    <StructLayout(LayoutKind.Sequential)> _
    Private Structure structSize
      Public cx As Integer
      Public cy As Integer
    End Structure

    <StructLayout(LayoutKind.Sequential)> _
    Private Structure WINDOWPOS
      Public hWnd As IntPtr
      Public hWndInsertAfter As IntPtr
      Public x As Integer
      Public y As Integer
      Public cx As Integer
      Public cy As Integer
      Public flags As Integer
    End Structure

    Private Declare Function GetCursorPos Lib "user32.dll" ( _
        ByRef lpPoint As POINT _
        ) As Boolean

    Private Declare Function ScreenToClient Lib "user32.dll" ( _
        ByVal hWnd As IntPtr, _
        ByRef lpPoint As POINT _
        ) As Boolean

    Private Declare Function GetWindowRect Lib "user32.dll" ( _
        ByVal hwnd As IntPtr, _
        ByRef lpRect As RECT _
        ) As Long

    Private Declare Function GetDC Lib "user32.dll" ( _
        ByVal hWnd As IntPtr _
        ) As IntPtr

    Private Declare Function GetParent Lib "user32.dll" ( _
        ByVal hWnd As IntPtr _
        ) As IntPtr

    Private Declare Function ReleaseDC Lib "user32.dll" ( _
        ByVal hWnd As IntPtr, _
        ByVal hdc As IntPtr _
        ) As Integer

		Private Declare Unicode Function GetTextExtentPoint32 Lib "gdi32.dll" Alias "GetTextExtentPoint32W" ( _
					ByVal hDC As IntPtr, _
					ByVal lpsz As String, _
					ByVal cbString As Integer, _
					ByRef lpSize As structSize _
		) As Integer

		Private Declare Unicode Function SelectObject Lib "gdi32.dll" (ByVal hDC As IntPtr, ByVal hObject As IntPtr) As IntPtr

		Private Declare Unicode Function SendMessage Lib "user32.dll" Alias "SendMessageW" ( _
		 ByVal hWnd As IntPtr, _
		 ByVal msg As Integer, _
		 ByVal wParam As Integer, _
		 ByVal lParam As Integer _
		 ) As IntPtr

    Private Declare Unicode Function SendMessage Lib "user32.dll" Alias "SendMessageW" ( _
        ByVal hWnd As IntPtr, _
        ByVal msg As Integer, _
        ByVal wParam As Integer, _
        ByVal lParam As IntPtr _
        ) As Integer

    Private Declare Unicode Function SendMessage Lib "user32.dll" Alias "SendMessageW" ( _
        ByVal hWnd As IntPtr, _
        ByVal msg As Integer, _
        ByVal wParam As IntPtr, _
        ByVal lParam As Integer _
        ) As Integer

    Private Declare Unicode Function SendMessage Lib "user32.dll" Alias "SendMessageW" ( _
        ByVal hWnd As IntPtr, _
        ByVal msg As Integer, _
        ByVal wParam As IntPtr, _
        ByVal lParam As IntPtr _
        ) As Integer

    Private Declare Unicode Function SendMessage Lib "user32.dll" Alias "SendMessageW" ( _
        ByVal hWnd As IntPtr, _
        ByVal msg As Integer, _
        ByVal wParam As Integer, _
        <MarshalAs(UnmanagedType.LPTStr)> _
        ByVal lParam As String _
        ) As Integer

    Private Declare Function ReleaseCapture Lib "user32.dll" Alias "ReleaseCapture" ( _
        ) As Boolean

    Private Declare Function SetCapture Lib "user32.dll" Alias "SetCapture" ( _
        ByVal hWnd As IntPtr _
        ) As IntPtr

    Private Const LB_GETTEXT As Integer = &H189
    Private Const LB_GETTEXTLEN As Integer = &H18A
    Private Const LB_ITEMFROMPOINT As Integer = &H1A9
#End Region

    Public Sub New()
      MyBase.new()

      tt.InitialDelayTime = 0
      tt.ReshowDelayTime = 0
      m_Timer.Enabled = False
      m_Timer.Interval = 20
    End Sub

    Public ReadOnly Property ToolTipToUse() As CToolTip
      Get
        If m_ExternalTT Is Nothing Then
          Return tt
        Else
          Return m_ExternalTT
        End If
      End Get
    End Property

    Public Property HotItem() As Integer
      Get
        Return m_HotItem
      End Get
      Set(ByVal Value As Integer)
        m_HotItem = Value
        If Not bAutoGetIndex Then UpdateToolTip(CShort(m_HotItem))
      End Set
    End Property

    Protected Overrides Sub WndProc(ByRef m As System.Windows.Forms.Message)
      Const SWP_HIDEWINDOW As Integer = &H80
      Const WM_MOUSEMOVE As Integer = &H200
      Const WM_WINDOWPOSCHANGED As Integer = &H47

      MyBase.WndProc(m)

      Select Case m.Msg
        Case WM_WINDOWPOSCHANGED
          If bUseTimer Then
            ' stop the timer if we got hidden
            Dim wp As WINDOWPOS

            wp = CType(Marshal.PtrToStructure(m.LParam, wp.GetType()), WINDOWPOS)
            If (wp.flags And SWP_HIDEWINDOW) = SWP_HIDEWINDOW Then
              m_Timer.Stop()
              m_Timer.Enabled = False
            End If
          End If
          UpdateToolTip(-1)

        Case WM_MOUSEMOVE
          If Not bUseTimer Then OnMouseMove(m)
      End Select
    End Sub

    Private Sub OnMouseMove(ByRef m As System.Windows.Forms.Message)
      Dim index As Short = CShort(Me.HotItem)
      Dim tmp As Integer = SendMessage(m.HWnd, LB_ITEMFROMPOINT, IntPtr.Zero, m.LParam)
      Dim insideClientArea As Boolean = (CShort(tmp >> 16) = 0)
			If bAutoGetIndex Then index = CShort(tmp And &HFFFF)

			If (Not insideClientArea) OrElse (index = ListBox.NoMatches) Then
				ReleaseCapture()
				UpdateToolTip(-1)
			Else
				SetCapture(Me.Handle)
				UpdateToolTip(index)
			End If
    End Sub

    Public Overloads Sub AssignHandle(ByVal handle As IntPtr, ByRef ExtToolTip As CToolTip, ByVal UseTimer As Boolean, ByVal AdjustTools As Boolean, ByVal AutoGetIndex As Boolean)
      m_ExternalTT = ExtToolTip
      bUseTimer = UseTimer
      bAdjustTools = AdjustTools
      bAutoGetIndex = AutoGetIndex

      Me.AssignHandle(handle)

      If bUseTimer Then
        m_Timer.Enabled = True
        m_Timer.Start()
      End If
    End Sub

    Public Overrides Sub CreateHandle(ByVal cp As System.Windows.Forms.CreateParams)
      MyBase.CreateHandle(cp)

      If Me.m_ExternalTT Is Nothing Then
        tt.Create(Control.FromHandle(Me.Handle))
        tt.Activate()
      End If
			Me.ToolTipToUse.AddTool(Me.Handle, "", , , bAdjustTools)
    End Sub

    Public Overrides Sub DestroyHandle()
      Me.ToolTipToUse.DelTool(Me.Handle)
      If Me.m_ExternalTT Is Nothing Then tt.Destroy()

      MyBase.DestroyHandle()
    End Sub

    Public Overloads Overrides Sub ReleaseHandle()
      Me.ToolTipToUse.DelTool(Me.Handle)
      If Me.m_ExternalTT Is Nothing Then tt.Destroy()

      MyBase.ReleaseHandle()
    End Sub

    Protected Overrides Sub OnHandleChange()
      Me.ToolTipToUse.DelTool(Me.Handle)
      If Me.m_ExternalTT Is Nothing Then
        tt.Destroy()
        tt.Create(Control.FromHandle(Me.Handle))
      End If
      Me.ToolTipToUse.AddTool(Me.Handle, "", , , bAdjustTools)
    End Sub


    Private Sub UpdateToolTip(ByVal index As Short)
			Const WM_GETFONT As Integer = &H31

			If index = ListBox.NoMatches Then
				' remove tooltip
				Me.ToolTipToUse.Text(Me.Handle) = ""
				Me.ToolTipToUse.Hide(Me.Handle)
			ElseIf bUseTimer OrElse (iLastIndex <> index) Then
				' get the item's text...
				Dim s As String = ""
				Dim sLen As Integer = SendMessage(Me.Handle, LB_GETTEXTLEN, CInt(index), IntPtr.Zero)

				If sLen > 0 Then
					s = Space(sLen)
					SendMessage(Me.Handle, LB_GETTEXT, CInt(index), s)

					' and calc its size
					Dim hDC As IntPtr = GetDC(Me.Handle)
					Dim hFont As IntPtr = SendMessage(Me.Handle, WM_GETFONT, 0, 0)
					Dim hPrevFont As IntPtr = SelectObject(hDC, hFont)
					Dim size As structSize
					GetTextExtentPoint32(hDC, s, s.Length, size)
					SelectObject(hDC, hPrevFont)
					ReleaseDC(Me.Handle, hDC)

					Dim rcWnd As RECT
					GetWindowRect(Me.Handle, rcWnd)
					If Not (rcWnd.right - rcWnd.left < size.cx) Then
						' all fine, we're smaller than the listbox
						s = ""
					End If
				End If

				' update tooltip
				Me.ToolTipToUse.Text(Me.Handle) = s
				If s = "" Then
					Me.ToolTipToUse.Hide(Me.Handle)
				Else
					Me.ToolTipToUse.Show(Me.Handle)
				End If
			End If

			If Not bUseTimer Then iLastIndex = index
    End Sub

    Private Sub m_timer_Tick(ByVal sender As Object, ByVal e As System.EventArgs) Handles m_Timer.Tick
      Dim pt As Point
      GetCursorPos(pt)
      ScreenToClient(Me.Handle, pt)

      Dim index As Short = CShort(Me.HotItem)
      Dim tmp As Integer = SendMessage(Me.Handle, LB_ITEMFROMPOINT, IntPtr.Zero, pt.ToInt32)
      Dim insideClientArea As Boolean = (CShort(tmp >> 16) = 0)
      If bAutoGetIndex Then index = CShort(tmp And &HFFFF)

      If (Not insideClientArea) OrElse (index = ListBox.NoMatches) Then
        UpdateToolTip(-1)
      Else
        UpdateToolTip(index)
      End If
    End Sub
  End Class




  Public Class ToolTipComboBox
    Inherits System.Windows.Forms.ComboBox

    Private LBox As New ToolTipListBox
    Private tt As New CToolTip

#Region "Unmanaged Code"
    Private Const WM_CTLCOLORLISTBOX As Integer = &H134
#End Region

    Public Sub New()
      MyBase.New()

      tt.InitialDelayTime = 0
      tt.ReshowDelayTime = 0
      If Me.DrawMode = DrawMode.Normal Then Me.DrawMode = DrawMode.OwnerDrawFixed
    End Sub

    Protected Overrides Sub WndProc(ByRef m As System.Windows.Forms.Message)
      If m.Msg = WM_CTLCOLORLISTBOX Then
        If LBox.Handle.Equals(IntPtr.Zero) Then
          ' creating a tooltip control with the listbox as parent will run us into trouble
          ' we also can't use SetCapture/ReleaseCapture, so use the timer instead
          LBox.AssignHandle(m.LParam, tt, True, True, False)
        End If
      End If

      MyBase.WndProc(m)
    End Sub

    Private Sub ToolTipComboBox_HandleCreated(ByVal sender As Object, ByVal e As System.EventArgs) Handles MyBase.HandleCreated
      If tt.Handle.Equals(IntPtr.Zero) Then
        tt.Create(CType(sender, Control))
        tt.Activate()
      End If
    End Sub

    Private Sub ToolTipComboBox_DrawItem(ByVal sender As Object, ByVal e As System.Windows.Forms.DrawItemEventArgs) Handles MyBase.DrawItem
      With CType(sender, ToolTipComboBox)
        If (e.State And DrawItemState.Selected) = DrawItemState.Selected Then
          If Me.DroppedDown Then
            ' this is the only chance to get the index
            LBox.HotItem = e.Index
          End If
        End If

        ' now do what's actually Windows' task
        e.DrawBackground()
        If (e.Index >= 0) And (e.Index <= Me.Items.Count) Then
          Dim br As New SolidBrush(e.ForeColor)
          Dim sz As SizeF = e.Graphics.MeasureString(CStr(Me.Items.Item(e.Index)), e.Font)
          Dim rc As New RectangleF(e.Bounds.X, e.Bounds.Y, sz.Width, sz.Height)
          e.Graphics.DrawString(CStr(Me.Items.Item(e.Index)), e.Font, br, rc)
        End If
        If ((e.State And DrawItemState.Focus) = DrawItemState.Focus) And ((e.State And DrawItemState.NoFocusRect) = 0) Then
          e.DrawFocusRectangle()
        End If
      End With
    End Sub

    Private Sub ToolTipComboBox_MeasureItem(ByVal sender As Object, ByVal e As System.Windows.Forms.MeasureItemEventArgs) Handles MyBase.MeasureItem
      With CType(sender, ToolTipComboBox)
        ' now do what's actually Windows' task
        If (e.Index >= 0) And (e.Index <= Me.Items.Count) Then
          Dim sz As Size = e.Graphics.MeasureString(CStr(Me.Items.Item(e.Index)), Me.Font).ToSize
          e.ItemHeight = sz.Height
          e.ItemWidth = sz.Width
        End If
      End With
    End Sub
  End Class
End Namespace